C MANDELM.F77
C
C Parallel Fortran v. 2.0: Copyright (c) 1988 3L Ltd
C Example program: Mandelbrot set evaluation and display
C
C Master task
C JF 10-May-88
C
C The application
C ---------------
C
C The application consists of two tasks:
C
C (1) MANDELM (this file).   This is the master task, and runs in the
C     root transputer.
C (2) MANDELW.   This is the worker task, and runs in all the other
C     transputers of the net.
C
C The flood configurer, FCONFIG, can be used to produce an executable
C file which will automatically distribute the worker tasks across an
C arbitrary network and route work packets from the master to the
C workers.
C
C It is also possible to run the application in a single transputer.
C This will work automatically if the application is configured using
C FCONFIG.  Alternatively, a static single-transputer configuration
C could be built by hand, using CONFIG.  A suitable configuration file
C may be found in MANDEL.CFG.
C
C As well as various routines from the Parallel Fortran run-time library,
C MANDELM must be linked with the CGA primitives module, CGA.BIN.
C A file MANDELM.LNK is supplied, which may be used to link MANDELM,
C like this:
C
C    LINKT @MANDELM.LNK,MANDELM.B4
C
C Functions of the tasks
C ----------------------
C
C MANDELM is told by the user which part of the Mandelbrot set to
C evaluate.   It then breaks this up into 100 packets, and sends them
C to the network of MANDELW's.   As the results from each return, they
C displayed on the PC's screen.
C
C The format of the data packet which is sent by MANDELM to MANDELW is
C defined in COMMAND.INC, and the format of the data packet sent back to
C MANDELM by MANDELW is defined in RESULTS.INC.
C
C Internals of MANDELM
C --------------------
C
C The task contains three threads.
C
C (1) The MAIN thread.
C This runs in the main program.  First, it intialises the other two
C threads and the integer ICHAN as an internal channel.  Then it goes
C into a loop, once round for each Mandelbrot display.  For each, it
C gets instructions from the user, and then sends to the SEND thread
C details of the part of the Mandelbrot set to be evaluated, using the
C ICHAN channel.  It keeps track of completed work by examining
C TALLY_DONE, which is incremented by RECEIVE every time a RESULTS
C packet is displayed; whenever it notices that TALLY_DONE has changed,
C it updates the PC's display; and when TALLY_DONE reaches 100, MAIN
C knows that the display is complete. 
C
C (2) The SEND thread.
C This waits until input arrives through the ICHAN channel, telling
C it what part of the Mandelbrot set to display.   When it arrives,
C this input goes straight into the first parts of the COMMAND data
C packet, whose format is defined in COMMAND.INC.   SEND then breaks
C the job into 100 small jobs.   For each, it adds the details to the
C COMMAND packet, and uses the F77_NET_SEND function to send it off
C to the network of MANDELW's. Notice that the SEND thread does not
C specify WHICH worker task is to do any particular job;  this is
C decided by the network of router tasks.
C
C (3) The RECEIVE thread.
C This simply waits till a packet arrives from the network of MANDELW's
C and then displays it.   Each packet contains all the necessary
C information to display it, so RECEIVE does not need to keep track of
C which packet is which.   Every time it does a display, RECEIVE
C increments TALLY_DONE, so that MAIN can tell when the whole display
C is complete.
C
C
C The SEND thread runs in this subroutine.
C
      SUBROUTINE SEND (X_INCREMENT, Y_INCREMENT)
      IMPLICIT NONE
      INTEGER X_INCREMENT, Y_INCREMENT
C
      INCLUDE 'COMMAND.INC'
      INCLUDE 'CGA.INC'
      INCLUDE 'CHAN.INC'
      INCLUDE 'NET.INC'
C
C     Channel to communicate with MAIN
      COMMON /INTERNAL_CHANNEL/ ICHAN
      INTEGER ICHAN
C
      INTEGER ICHANADDR, X, Y
C
      ICHANADDR = F77_CHAN_ADDRESS (ICHAN)
C
100   CONTINUE
C
C        Wait for instructions from MAIN thread
         CALL F77_CHAN_IN_MESSAGE (12, COMMAND, ICHANADDR)
C        Send off the packets to be done.   Each includes the top-left
C        and bottom-right corners of the area to do.   This both tells
C        the worker task what value to generate and identifies the
C        RESULTS packet when it arrives in the RECEIVE thread (since
C        there's not guarantee that the results will arrive in the
C        same order that the commands were sent out.)
         DO X = 0, CGA_LORES_XMAX-1, X_INCREMENT
            C_TLX = X
            C_BRX = X + X_INCREMENT-1
            DO Y = 0, CGA_YMAX-1, Y_INCREMENT
               C_TLY = Y
               C_BRY = Y + Y_INCREMENT-1
C              Send off the next packet
               CALL F77_NET_SEND (28, COMMAND, .TRUE.)
            END DO
         END DO
C
      GOTO 100
C
      END
C
C
C The RECEIVE thread runs in this subroutine
C
      SUBROUTINE RECEIVE (TALLY_DONE)
      IMPLICIT NONE
      INTEGER TALLY_DONE
C
      INCLUDE 'CGA.INC'
      INCLUDE 'NET.INC'
      INCLUDE 'RESULTS.INC'
C
      INTEGER LEN, X, Y, I, N, COLOUR
      LOGICAL COMPLETE
C
      INTEGER THRESH1, THRESH2, THRESH3
      COMMON /THRESHOLDS/ THRESH1, THRESH2, THRESH3
C
100   CONTINUE
C
C        Thread will wait here till a packet arrives
         CALL F77_NET_RECEIVE (LEN, RESULTS, COMPLETE)
C        The RESULTS packet includes the coordinates of the top-left
C        and bottom-right corners of the data, so we know where to
C        display them.

         I = 1
         DO Y = R_TLY, R_BRY
            DO X = R_TLX, R_BRX
               N = ICHAR (R_COUNTS(I:I))
               I = I + 1
C              Received 0 means 1; received 255 means 256; etc
               N = N + 1
C              Decide on colour from thresholds and display
               IF (N .GT. THRESH3) THEN
                  COLOUR = 3
               ELSE IF (N .GT. THRESH2) THEN
                  COLOUR = 2
               ELSE IF (N .GT. THRESH1) THEN
                  COLOUR = 1
               ELSE
                  COLOUR = 0
               ENDIF
               CALL CGA_LORES_PLOT (X, Y, COLOUR)
            END DO
         END DO
C        Increment tally of packets displayed
         TALLY_DONE = TALLY_DONE + 1
C
      GOTO 100
C
      END
C
C
C The MAIN thread runs here
C
      PROGRAM MAIN
      IMPLICIT NONE
C
      INCLUDE 'THREAD.INC'
      INCLUDE 'CHAN.INC'
      INCLUDE 'CGA.INC'
C
      INTEGER WS_SIZE
      PARAMETER (WS_SIZE=2500)
C
      EXTERNAL SEND, RECEIVE
C
C     Channel for communication with SEND thread
      INTEGER ICHAN
      COMMON /INTERNAL_CHANNEL/ ICHAN
C
C     Common block for communication with RECEIVE thread
      INTEGER THRESH1, THRESH2, THRESH3
      COMMON /THRESHOLDS/ THRESH1, THRESH2, THRESH3
C
C     Definition of shape and number of packets
      INTEGER PACKETS, X_INCREMENT, Y_INCREMENT
      PARAMETER (PACKETS     = 100,
     1           X_INCREMENT = (CGA_LORES_XMAX+1)/10,
     2           Y_INCREMENT = (CGA_YMAX+1)/10        )
C
      INTEGER ICHANADDR, SEND_WS(WS_SIZE), RECEIVE_WS(WS_SIZE),
     1        TALLY_DONE, PREVIOUS_TALLY
      REAL USER_INPUT(3), RANGE, XCOORD, YCOORD, GAP
      EQUIVALENCE (USER_INPUT(1), XCOORD),
     1            (USER_INPUT(2), YCOORD),
     2            (USER_INPUT(3), GAP)
C
C     Text mode, clear screen and sign on.
      CALL VIDEO_MODE (MONO_80COL_TEXT_MODE)
      PRINT *,'Parallel Fortran v. 2.0'
      PRINT *,'Example program: Mandelbrot set evaluation and display'
      PRINT *,'Copyright (c) 1988 3L Ltd'
      PRINT *,'This program requires a Colour Graphics Adaptor (CGA)'
      PRINT *
C
C     Initialise the channel for communication with SEND
      ICHANADDR = F77_CHAN_ADDRESS (ICHAN)
      CALL F77_CHAN_INIT (ICHANADDR)
C
C     Start the other threads 
      TALLY_DONE = 0
      CALL F77_THREAD_START (SEND, SEND_WS, WS_SIZE*4,
     1                       F77_THREAD_URGENT,
     2                       2, X_INCREMENT, Y_INCREMENT)
      CALL F77_THREAD_START (RECEIVE, RECEIVE_WS, WS_SIZE*4,
     1                       F77_THREAD_URGENT,
     2                       1, TALLY_DONE)
C
100   CONTINUE
C
C        This will ensure that no other threads are using the Run-
C        Time Library (in fact, in this case they won't be, but I
C        have done it here as an example...)
         CALL F77_THREAD_USE_RTL
         PRINT '(''Input X coordinate:'',$)'
         READ *,XCOORD
         PRINT '(''Input Y coordinate:'',$)'
         READ *,YCOORD
         PRINT '(''Input Y range:     '',$)'
         READ *,RANGE
         GAP = RANGE / (CGA_YMAX+1)
         YCOORD = YCOORD + RANGE
C
         PRINT '(''Threshold 1: '',$)'            
         READ *,THRESH1
         PRINT '(''Threshold 2: '',$)'            
         READ *,THRESH2
         PRINT '(''Threshold 3: '',$)'            
         READ *,THRESH3
C
C        Finished with Run-Time Library - release it
         CALL F77_THREAD_FREE_RTL
C
C        Into graphics (CGA low-resolution) mode
         CALL VIDEO_MODE (CGA_LORES_GRAPHICS_MODE)
C
C        Reset count of finished packets to zero: RECEIVE
C        will count it back up as packets come back
         TALLY_DONE = 0
C
C        Send instructions to SEND
         CALL F77_CHAN_OUT_MESSAGE (12, USER_INPUT, ICHANADDR)
C
C        Until all the packets have been done, just keep
C        updating the display when necessary.
         PREVIOUS_TALLY = 0
         DO WHILE (TALLY_DONE .LT. PACKETS)
            DO WHILE (TALLY_DONE .EQ. PREVIOUS_TALLY)
               CALL F77_THREAD_DESCHEDULE
            END DO
            CALL CGA_UPDATE
            PREVIOUS_TALLY = TALLY_DONE
         END DO
C        Extra display update, in case...
         CALL CGA_UPDATE
C
C        Once again, wait for the Run-Time Library; then beep and
C        wait till the user hits return
         CALL F77_THREAD_USE_RTL
         PRINT '(A1)', 7
         READ (*,*)
         CALL F77_THREAD_FREE_RTL
C
C        Clear the screen and text mode again
         CALL VIDEO_MODE (MONO_80COL_TEXT_MODE)
C
      GOTO 100
C
      END

